/*
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.0 (the "License"); you may not use this file except in compliance 
 * with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License. 
 * 
 * The Original Code is the roamd Version 0.1 code,
 * released on or about December 4, 1998.
 * 
 * The Initial Developer of the Original Code is John Richard Stracke,
 * Jr.  Portions created by John Richard Stracke, Jr. are Copyright (C)
 * John Richard Stracke, Jr. All Rights Reserved. 
 * 
 * Contributor(s): ______________________________________. 
 */
/*
 * Implements the port function(s) defined in port.h.
 *
 * $Revision: 1.5 $
 * $Author: francis $
 * $Date: 1998/12/08 23:29:39 $
 *
 * $Log: port.c,v $
 * Revision 1.5  1998/12/08 23:29:39  francis
 * Adding <sys/types.h> before <netinet/in.h>; needed on some platforms
 *
 * Revision 1.4  1998/12/06 20:02:22  francis
 * Adding ident lines
 *
 * Revision 1.3  1998/12/04 22:03:20  francis
 * Adding license terms to the source files.
 * Adding tryCrypt and tryCrypt.o to the clean target.
 * Deleting tryCrypt from the repository.
 * Adding Index.html and license.html.  Linking to license.html from
 * readme.html.  A bit of cosmetic work in readme.html.
 *
 * Revision 1.2  1998/12/04 20:01:03  francis
 * Adding documentation headers and comments
 *
 */
#include "ident.h"
IDENTIFY(ROAMD_PORT_C,"$Id: port.c,v 1.5 1998/12/08 23:29:39 francis Exp $")

#include <sys/types.h>
#include <netdb.h>
#include <netinet/in.h>

#include "port.h"

/*
 * Used to convert a human-convenient port string into a port number (in
 *  host order).
 */
int getPort(const char* portOrService,
            int defaultPort)
{
  /* Can we convert portOrService to an integer? */
  int p=atoi(portOrService);
  /* If so, then p came back nonzero */
  if (p!=0)
    return p;

  /* But maybe it was "0"--which is a useful value (INADDR_ANY) */
  if (!strcmp(portOrService,"0"))
    return 0;

  {
    /* So it's a service name; look it up */
    struct servent* serv=getservbyname(portOrService,"tcp");

    /* If it doesn't exist, return the default */
    if (!serv)
      return defaultPort;

    /* Convert the service's port to host order */
    return ntohs(serv->s_port);
  }
}
