/*
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.0 (the "License"); you may not use this file except in compliance 
 * with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License. 
 * 
 * The Original Code is the roamd Version 0.1 code,
 * released on or about December 4, 1998.
 * 
 * The Initial Developer of the Original Code is John Richard Stracke,
 * Jr.  Portions created by John Richard Stracke, Jr. are Copyright (C)
 * John Richard Stracke, Jr. All Rights Reserved. 
 * 
 * Contributor(s): ______________________________________. 
 */
#ifndef ROAMD_PORT_H
#define ROAMD_PORT_H

#include "ident.h"
IDENTIFY(ROAMD_PORT_H,"$Id: port.h,v 1.4 1998/12/06 20:02:22 francis Exp $")

/*
 * Declares functions for converting a port string (either a service name
 *  or a port represented as a string) to a port number.
 *
 * $Revision: 1.4 $
 * $Author: francis $
 * $Date: 1998/12/06 20:02:22 $
 *
 * $Log: port.h,v $
 * Revision 1.4  1998/12/06 20:02:22  francis
 * Adding ident lines
 *
 * Revision 1.3  1998/12/04 22:03:20  francis
 * Adding license terms to the source files.
 * Adding tryCrypt and tryCrypt.o to the clean target.
 * Deleting tryCrypt from the repository.
 * Adding Index.html and license.html.  Linking to license.html from
 * readme.html.  A bit of cosmetic work in readme.html.
 *
 * Revision 1.2  1998/12/04 20:01:03  francis
 * Adding documentation headers and comments
 *
 */

/*
 * Used to convert a human-convenient port string into a port number (in
 *  host order).  Called by procargs() when it gets a "--port".
 *
 * If portOrService is "0", returns 0 (i.e., ask bind(2) to pick a port--
 *  used for development purposes, when you don't want to wait for
 *  the port to time out).
 * If portOrService is the string rep of an integer, returns that integer.
 * If portOrService is a TCP service name from /etc/services (or wherever),
 *  returns the port number of that service (in host order).
 * Otherwise, returns defaultPort.
 */
int getPort(const char* portOrService,
            int defaultPort);

#endif /*ROAMD_PORT_H*/
