/*
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.0 (the "License"); you may not use this file except in compliance 
 * with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License. 
 * 
 * The Original Code is the roamd Version 0.1 code,
 * released on or about December 4, 1998.
 * 
 * The Initial Developer of the Original Code is John Richard Stracke,
 * Jr.  Portions created by John Richard Stracke, Jr. are Copyright (C)
 * John Richard Stracke, Jr. All Rights Reserved. 
 * 
 * Contributor(s): ______________________________________. 
 */
/*
 * Implements the main body of the roamd program.
 *
 * This module is pretty hairy; it grew more than it was designed.
 *
 * $Revision: 1.19 $
 * $Author: francis $
 * $Date: 1999/06/21 04:06:49 $
 *
 * $Log: roamd.c,v $
 * Revision 1.19  1999/06/21 04:06:49  francis
 * Fixing a bug introduced by the last checkin
 *
 * Revision 1.18  1999/06/21 04:04:37  francis
 * Making the bookmarks file be text/html
 *
 * Revision 1.17  1999/03/21 01:52:57  francis
 * Fixing some things from the last checkin:
 * In users.c, having eachPass() return N (this stuff was copied over
 * by hand from a version done on Solaris [where I had Purify], and I
 * left something out).  Also adding some debugs that helped me find
 * the problem.
 * In roamd.c, adding --noFork to the usage(); also fixing its procargs()
 * clause (before, it fell through to the "unrecognized option" part).
 * In Makefile, fixing the new purify part of clean.
 *
 * Revision 1.16  1999/03/21 01:23:58  francis
 * Fixing some bugs spotted by Purify, including a file descriptor leak.
 * Adding fork-per-connection (disable with the --noFork option).
 *
 * Revision 1.15  1999/01/11 16:27:57  francis
 * Adding a really severe diag in expect(), to track every character.
 * Controlled by DEBUG_EXCESSIVE.
 *
 * Revision 1.14  1999/01/11 16:13:00  francis
 * Adding extra diags in expect()
 *
 * Revision 1.13  1998/12/17 20:07:54  francis
 * Adding SO_REUSEADDR and ignoring SIGPIPE.
 *
 * Revision 1.12  1998/12/13 06:44:50  francis
 * Adding includes to find malloc() on more platforms
 *
 * Revision 1.11  1998/12/11 16:16:07  francis
 * Fixing bug where I was setting addr.sin_family and *then* zeroing addr.
 * Thanks to Philip Flip Kromer <flip@physics.utexas.edu> for catching this.
 * Awfully embarrassing...
 *
 * Revision 1.10  1998/12/08 06:47:35  francis
 * Fixing another Solaris 2.6 x86 bug
 *
 * Revision 1.9  1998/12/08 06:42:13  francis
 * Version fixing Solaris 2.6 x86 bugs found by Teknix <teknix@alloy.net>
 *
 * Revision 1.8  1998/12/06 20:02:22  francis
 * Adding ident lines
 *
 * Revision 1.7  1998/12/06 19:41:27  francis
 * Adding MOVE support.
 * Adding strip target to Makefile.
 *
 * Revision 1.6  1998/12/04 22:03:21  francis
 * Adding license terms to the source files.
 * Adding tryCrypt and tryCrypt.o to the clean target.
 * Deleting tryCrypt from the repository.
 * Adding Index.html and license.html.  Linking to license.html from
 * readme.html.  A bit of cosmetic work in readme.html.
 *
 * Revision 1.5  1998/12/04 20:51:09  francis
 * Removing the shutdown(srvr) in cleanup(), since it's not needed
 *
 * Revision 1.4  1998/12/04 20:50:12  francis
 * Fixing password authentication to make sure that the username in the
 * credentials is the owner of the requested file.
 *
 * Revision 1.3  1998/12/04 20:44:22  francis
 * Documenting roamd.c better
 *
 * Revision 1.2  1998/12/04 20:01:03  francis
 * Adding documentation headers and comments
 *
 */

#include "ident.h"
IDENTIFY(ROAMD_ROAMD_C,"$Id: roamd.c,v 1.19 1999/06/21 04:06:49 francis Exp $")

#include <sys/param.h>

#ifndef NAME_MAX
#define NAME_MAX MAXNAMELEN
#endif

#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <malloc.h>
#include <stdlib.h>

#include "passwd.h"
#include "users.h"

int forkPerConnection=1;

#ifdef DEBUG
int suppressExpectIllegals=1;
#endif

/* The name under which the program was invoked */
const char* progname="roamd";

/* The directory under which config data lives (currently, there's
 *  nothing there but the users directory).  This is the default;
 *  it gets overridden by the --configDir option.
 */
const char* configDir="/etc/roamd";

/* The directory under which user profiles are stored.  Built from
 *  configDir after procargs() is called.
 */
const char* userDir=0;

/* Clean up and exit */
void cleanup();
/* Signal handler */
void signalHandler(int sig);
/* Initialize the signal handlers */
void initSignals(void);

/* Process the command-line arguments */
void procargs(int argc, const char* argv[]);

/* Process an incoming connection */
void processRequest(int sock);

/* The server socket */
static int srvr=-1;
/* The default port to use */
static const int defaultPort=7500;
/* The port we're binding to (if 0, bind() will pick one) */
static int port;

/* The user names returned by initUsers() (see users.h).  A 0-terminated
 *  array of 0-terminated strings.
 */
const char** userNames=0;

/* Report that the function named "call" returned an error (specified in
 *  errno); if exitCode!=0, exit(exitCode)
 */
void reportError(const char* call,
                 int exitCode);

/* Convert the given time_t into an HTTP time string */
const char* makeHttpTime(time_t t);
/* Parse the given HTTP time string into a time_t */
time_t parseHttpTime(const char* timeStr);

/* Test: does the filename match the pattern "li[0-9]*\.tmp"? */
int fileMatchesTmpPattern(const char* file);

/* Main program */
int main(int argc, const char* argv[])
{
  initSignals();

  /* I'd prefer to do this when port is declared, but the compiler
   *  doesn't like that; it doesn't believe that const means const.
   */
  port=defaultPort;

  /* Process the command-line arguments */
  procargs(argc,argv);

  /* Build userDir from configDir */
  {
    char _userDir[MAXPATHLEN];
    sprintf(_userDir,"%s/users",configDir);
    userDir=strdup(_userDir);
  }

  /* Get the array of user names */
  userNames=initUsers(userDir);
  if (!userNames)
    exit(11);

  /* Create the server socket */
  srvr=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
  if (srvr<0)
    reportError("socket",1);

  {
    int optval=1;
    if (setsockopt(srvr,SOL_SOCKET,SO_REUSEADDR,
                   (void*)&optval,sizeof(optval))<0)
      {
        int tmp=errno;
        fprintf(stderr,"%s: setsockopt(SO_REUSEADDR): %s\n",
                progname,
                strerror(tmp));
        fprintf(stderr,"%s: next execution of %s may have to wait for TIME_WAIT sockets to expire\n",
                progname,progname);
      }
  }

  {
    struct sockaddr_in addr;
    memset(&addr,0,sizeof(addr));
    addr.sin_family=AF_INET;
    addr.sin_port=htons(port);
    /* Accept connections on all interfaces */
    addr.sin_addr.s_addr=INADDR_ANY;

    /* Bind the server socket */
    if (bind(srvr,(struct sockaddr*)&addr,sizeof(addr))<0)
      reportError("bind",2);

    /* If we didn't specify a port, find out what bind() gave us
     *  and report it.
     */
    if (!port)
      {
        int addrlen=sizeof(addr);
        if (getsockname(srvr,(struct sockaddr*)&addr,&addrlen)<0)
          reportError("getsockname",3);
        printf("%s bound to port %d\n",
               progname,
               (int)(ntohs(addr.sin_port))
               );
      }

    /* Start listening on the port */
    if (listen(srvr,4)<0)
      reportError("listen",5);
  }

  /* Main loop */
  while (1)
    {
      struct sockaddr_in addr;
      int addrlen=sizeof(addr);

      /* Wait for an incoming connection */
      int newsock=accept(srvr,(struct sockaddr*)&addr,&addrlen);
      if (newsock<0)
        reportError("accept",6);

#ifdef DEBUG
      printf("Incoming connection\n");
#endif

      if ((!forkPerConnection) || (fork()>0))
        {
          /* Handle it */
          processRequest(newsock);

          /* Close down the connection.  If we can't, we'll report an error,
           *  but not exit--it's not fatal until the fds fill up.
           */
          if (shutdown(newsock,2)<0)
            reportError("shutdown",0);
          if (close(newsock)<0)
            reportError("close",0);

          if (forkPerConnection)
            exit(0);
        }
    }

  cleanup();
  return 0;
}

/* Report the program's usage */
void usage(void)
{
  fprintf(stderr,"Usage: %s [--port <port>] [--configDir <dir>] [--noFork]\n",progname);
  exit(10);
}

/* Process the command-line arguments */
void procargs(int argc, const char* argv[])
{
  int i;

  /* Initialize progname */
  if (strchr(argv[0],'/'))
    progname=strrchr(argv[0],'/')+1;
  else
    progname=argv[0];

  i=1;
  while (i<argc)
    {
      const char* opt=argv[i++];
      if (!strcmp(opt,"--port"))
        {
          const char* arg=argv[i++];
          if (!arg)
            usage();

          /* Convert the port string or service name to a port */
          port=getPort(arg,defaultPort);
          continue;
        }

      if (!strcmp(opt,"--configDir"))
        {
          const char* arg=argv[i++];
          if (!arg)
            usage();

          configDir=arg;
          continue;
        }

      if (!strcmp(opt,"--testFile"))
        {
          const char* arg=argv[i++];
          if (!arg)
            usage();

          printf("fileMatchesTmpPattern(%s): %d\n",
                 arg,
                 fileMatchesTmpPattern(arg));
          exit(0);
        }

      if (!strcmp(opt,"--noFork"))
        {
          forkPerConnection=0;
          continue;
        }

      usage();
    }
}

/* Report that the function named "call" returned an error (specified in
 *  errno); if exitCode!=0, exit(exitCode)
 */
void reportError(const char* call,
                 int exitCode)
{
  int tmp=errno;
  fprintf(stderr,"%s: %s: %s\n",
          progname,
          call,
          strerror(tmp)
          );
  if (exitCode)
    {
      cleanup();
      exit(exitCode);
    }
}

/* Clean up and exit */
void cleanup()
{
  static int alreadyBusy=0;
  if (alreadyBusy)
    return;

  alreadyBusy++;
  
  if (close(srvr)<0)
    reportError("close",4);

  alreadyBusy--;
}

/* Initialize the signal handlers */
void initSignals(void)
{
  /* List of signals we want to catch.  Each is enclosed in an #ifdef
   *  for portability.
   */
  int signals[]={
#ifdef SIGINT
    SIGINT,
#endif
#ifdef SIGQUIT
    SIGQUIT,
#endif
#ifdef SIGILL
    SIGILL,
#endif
#ifdef SIGTRAP
    SIGTRAP,
#endif
#ifdef SIGIOT
    SIGIOT,
#endif
#ifdef SIGBUS
    SIGBUS,
#endif
#ifdef SIGFPE
    SIGFPE,
#endif
#ifdef SIGSEGV
    SIGSEGV,
#endif
#ifdef SIGPIPE
    SIGPIPE,
#endif
#ifdef SIGTERM
    SIGTERM,
#endif
#ifdef SIGSTKFLT
    SIGSTKFLT,
#endif
#ifdef SIGXCPU
    SIGXCPU,
#endif
#ifdef SIGXFSZ
    SIGXFSZ,
#endif
#ifdef SIGPWR
    SIGPWR
#endif
  };

  int i,N;
  for (i=0,N=(sizeof(signals)/sizeof(signals[0])); i<N; i++)
    signal(signals[i],signalHandler);

#ifdef SIGPIPE
  signal(SIGPIPE,SIG_IGN);
#endif
}

/* Signal handler */
void signalHandler(int sig)
{
  /* Try to clean up before exiting--but make sure that, if we get a
   *  second signal (i.e., we got another signal during cleanup), we
   *  just quit, no further questions.
   */

  static int alreadyBusy=0;
  if (!alreadyBusy)
    {
      alreadyBusy=1;
      fprintf(stderr,"%s: exiting with signal %d\n",
              progname,sig);
      cleanup();
    }
  exit(sig+128);
}

/* Enums representing legal HTTP methods. */
typedef enum {
  MethodInvalid=0,
  MethodGet,
  MethodHead,
  MethodPut,
  MethodDelete,
  MethodMove
} Method;

/* Get the length of the longest string in the array */
int maxLength(const char* choices[])
{
  int i=0;
  int max=0;

  if (!choices)
    return 0;

  while (choices[i])
    {
      int len=strlen(choices[i++]);
      if (len>max)
        max=len;
    }

  return max;
}

/* Read one byte from the socket.  Returns that byte, or <0 on error */
int read1(int sock)
{
  char ch;
  if (read(sock,&ch,1)<1)
    {
      reportError("read",0);
      return -1;
    }

  return ch;
}

/* Convenience function: does a strcmp() if caseSensitive is true,
 *  strcasecmp() otherwise.
 */
int strmaybecmp(const char* s1,
                const char* s2,
                int caseSensitive)
{
  return (caseSensitive ? strcmp : strcasecmp)(s1,s2);
}

/*
 * The heart of the finite state machine: read a string from the socket;
 *  if it's one of the strings in the choices array, return that string's
 *  index; if not, keep reading until you reach the terminator.
 *
 * If caseSensitive is true, then the string found must match a choices
 *  entry exactly; if not, then case is ignored.
 */
int expect(int sock,
           const char* choices[],
           char terminator,
           int caseSensitive)
{
  struct {
    char* data;
    int size,offset;
  } buff;
  int ch;

  if (!choices)
    return -1;

  buff.size=maxLength(choices);

  if (!buff.size)
    return -1;

  buff.data=malloc(buff.size+1);
  if (!buff.data)
    reportError("expect: malloc",7);

  buff.offset=0;

  while (buff.offset<buff.size)
    {
      ch=read1(sock);
      if (ch<0)
        {
          fprintf(stderr,"%s: expect: unexpected EOF\n",progname);
          free(buff.data);
          return -1;
        }

      if (ch=='\n')
        {
          fprintf(stderr,"%s: expect: unexpected \\n\n",progname);
          free(buff.data);
          return -2;
        }

#ifdef DEBUG_EXCESSIVE
      fprintf(stderr,"%s: expect: read '%c' (char #%d)\n",progname,ch,(int)ch);
#endif

      if (ch==terminator)
        break;
      buff.data[buff.offset++]=ch;
    }

  buff.data[buff.offset]=0;

  if (ch!=terminator)
    {
      ch=read1(sock);
      if (ch!=terminator)
        {
          fprintf(stderr,"%s: expect: badly terminated\n",progname);
          fprintf(stderr,"%s: expect: got \"%s\"\n",progname,buff.data);
          fprintf(stderr,"%s: expect: choices were:\n",progname);
          {
            int i;
            for (i=0; choices[i]; i++)
              fprintf(stderr,"\t\"%s\"\n",choices[i]);
          }
          free(buff.data);
          return -1;
        }
    }

  {
    int i;
    for (i=0; choices[i]; i++)
      {
        if (!strmaybecmp(buff.data,choices[i],caseSensitive))
          {
            free(buff.data);
            return i;
          }
      }
  }

#ifdef DEBUG
  if (!suppressExpectIllegals)
    {
      fprintf(stderr,"%s: expect: found illegal choice \"%s\"\n",progname,buff.data);
      fprintf(stderr,"%s: choices were:\n",progname);
      {
        int i;
        for (i=0; choices[i]; i++)
          fprintf(stderr,"\t\"%s\"\n",choices[i]);
      }
    }
#endif
  free(buff.data);
  return -1;
}

/* Read a method from the socket; return the corresponding Method
 *  (enum), or MethodInvalid if nothing legal is found.
 */
Method getMethod(int sock)
{
  static const char* methods[]={"GET","HEAD","PUT","DELETE","MOVE",0};
  int m=expect(sock,methods,' ',1);
  if (m<0)
    return MethodInvalid;

  return (Method)(m+1);
}

/* Convert the given Method enum to a string */
const char* methodName(Method method)
{
  switch (method)
    {
    case MethodInvalid: return "Invalid";
    case MethodGet: return "Get";
    case MethodHead: return "Head";
    case MethodPut: return "Put";
    case MethodDelete: return "Delete";
    case MethodMove: return "Move";
    default: return "Unknown";
    }
}

/* More efficient substitute for expect() with one argument (case
 *  sensitive)
 */
int expect1(int sock,
            const char* str)
{
  if (!str)
    return 0;

  while (*str)
    {
      int ch=read1(sock);
      if (ch<0)
        return -1;

      if (ch!=(*str))
        return -1;

      str++;
    }

  return 0;
}

/* Read a line from the socket.  If skipLWS is true, skips leading
 *  whitespace.
 */
int readLine(int sock,
             char* buff,
             int buffsize,
             int skipLWS)
{
  int offset;

  if (!buff)
    return -1;
  if (buffsize<0)
    return -1;

  for (offset=0; offset<(buffsize-1);)
    {
      int ch=read1(sock);
      if (ch<0)
        return -1;

      if (ch=='\n')
        break;

      if (skipLWS && isspace(ch))
        continue;
      skipLWS=0;

      buff[offset++]=ch;
    }

  buff[offset]=0;
  return offset;
}

/* Skip the rest of the line */
int skipLine(int sock)
{
  int ch=' ';
  while ((ch!='\n') && (ch>=0))
    ch=read1(sock);

  if (ch<0)
    return -1;
  else
    return 0;
}

/* Read a username from the socket.  If no legal username is found,
 *  returns 0.
 */
const char* getUser(int sock)
{
  int userNum=expect(sock,userNames,'/',1);
  if (userNum<0)
    return 0;

  return userNames[userNum];
}

const char* readUntil(int sock,
                      char terminator,
                      int maxLen)
{
  char* buff=malloc(maxLen+1);
  if (!buff)
    return 0;

  {
    int offset=0;
    while (offset<maxLen)
      {
        int ch=read1(sock);
        if (ch<0)
          {
            int tmp=errno;
            free(buff);
            errno=tmp;
            return 0;
          }

        if (ch==terminator)
          {
            buff[offset]=0;
            return buff;
          }

        buff[offset++]=ch;
      }

    fprintf(stderr,
            "%s: readUntil(): no terminator ('%c') found within maxLen (%d) bytes\n",
            progname,
            terminator,
            maxLen);
    free(buff);
    errno=EIO;
    return 0;
  }
}

/* Build the pathname of the requested user's requested file */
const char* getFilePath(const char* user,
                        const char* file)
{
  static char path[MAXPATHLEN];
  sprintf(path,"%s/%s/%s",
          userDir,
          user,
          file);
  return path;
}

#ifdef PRE_MOVE_GETFILE
/* Read a legal filename from the socket */
const char* getFile(int sock)
{
  /* These are the files used by Communicator 4.5's roaming access; no
   *  others are permitted.
   */
  static const char* files[]={
    "IMAP",
    "WRITETEST",
    "bookmarks",
    "certdb",
    "cookies",
    "filters",
    "globalhistory",
    "keydb",
    "liprefs",
    "pab.na2",
    "secmod",
    "signed",
    0
  };
  
  static const int numFiles=sizeof(files)/sizeof(files[0]);

  int fileNum=expect(sock,files,' ',1);
  if (fileNum<0)
    return 0;

  return files[fileNum];
}
#else

int fileMatchesTmpPattern(const char* file)
{
  const char* cur=file;
  if (!cur)
    {
#ifdef DEBUG
      fprintf(stderr,"fileMatchesTmpPattern(NULL)\n");
#endif
      return 0;
    }

 start:
  if ((*(cur++))!='l')
    {
#ifdef DEBUG
      fprintf(stderr,"fileMatchesTmpPattern(%s): '%c' instead of 'l'\n",file,cur[-1]);
#endif
      return 0;
    }
  if ((*(cur++))!='i')
    {
#ifdef DEBUG
      fprintf(stderr,"fileMatchesTmpPattern(%s): '%c' instead of 'i'\n",file,cur[-1]);
#endif
      return 0;
    }

 num:
  if (isdigit(*(cur++)))
    goto num;
  
 suffix:
  cur--;
  if ((*(cur++))!='.')
    {
#ifdef DEBUG
      fprintf(stderr,"fileMatchesTmpPattern(%s): '%c' instead of '.'\n",file,cur[-1]);
#endif
      return 0;
    }
  if ((*(cur++))!='t')
    {
#ifdef DEBUG
      fprintf(stderr,"fileMatchesTmpPattern(%s): '%c' instead of 't'\n",file,cur[-1]);
#endif
      return 0;
    }
  if ((*(cur++))!='m')
    {
#ifdef DEBUG
      fprintf(stderr,"fileMatchesTmpPattern(%s): '%c' instead of 'm'\n",file,cur[-1]);
#endif
      return 0;
    }
  if ((*(cur++))!='p')
    {
#ifdef DEBUG
      fprintf(stderr,"fileMatchesTmpPattern(%s): '%c' instead of 'p'\n",file,cur[-1]);
#endif
      return 0;
    }

  if (*cur)
    {
#ifdef DEBUG
      fprintf(stderr,"fileMatchesTmpPattern(%s): '%c' instead of 0\n",file,cur[-1]);
#endif
      return 0;
    }

  return 1;
}

int tmpFilesLegal(Method method)
{
  switch (method)
    {
    case MethodPut:
    case MethodDelete:
    case MethodMove:
       return 1;

    default:
       return 0;
    }
}

/* Read a legal filename from the socket */
const char* getFile(int sock,
                    int permitTmpFiles)
{
  /* These are the files used by Communicator 4.5's roaming access; no
   *  others are permitted.
   */
  static const char* files[]={
    "IMAP",
    "WRITETEST",
    "bookmarks",
    "certdb",
    "cookies",
    "filters",
    "globalhistory",
    "keydb",
    "liprefs",
    "pab.na2",
    "secmod",
    "signed",
    0
  };
  
  const char* file=readUntil(sock,' ',NAME_MAX);

  if (!file)
    return 0;

#ifdef DEBUG
  fprintf(stderr,"%s: getFile(): readUntil() returned \"%s\"\n",
          progname,file);
#endif

  if (permitTmpFiles && (fileMatchesTmpPattern(file)))
    {
      static char buff[NAME_MAX+1];
      strcpy(buff,file);
      free((void*)file); /* cast away const */
      return buff;
    }

  {
    int i;
    for (i=0; files[i]; i++)
      {
        if (!strcmp(file,files[i]))
          {
            free((void*)file); /* cast away const */
            return files[i];
          }
      }
  }

  free((void*)file);
  return 0;
}
#endif

/* Enum of the relevant headers */
typedef enum {
  HeaderInvalid=0,
  HeaderContentLength,
  HeaderIfModifiedSince,
  HeaderAuthorization,
  HeaderNewUri,
  HeaderContentDanger, /* Used for Content-* headers that we don't recognize */
  HeaderTerminator
} Header;

struct {
  char contentLength[40];
  char ifModifiedSince[1024];
  char authorization[1024];
  char newUri[1024];
} headerData={"","",""};

struct {
  const char* name;
  Header tag;
  char* string;
  int maxStringLen;
} headerDescs[]={
  {
    "Content-Length",
    HeaderContentLength,
    headerData.contentLength,
    sizeof(headerData.contentLength)
  },
  {
    "If-Modified-Since",
    HeaderIfModifiedSince,
    headerData.ifModifiedSince,
    sizeof(headerData.ifModifiedSince)
  },
  {
    "Authorization",
    HeaderAuthorization,
    headerData.authorization,
    sizeof(headerData.authorization)
  },
      {
        "New-uri",
        HeaderNewUri,
        headerData.newUri,
        sizeof(headerData.newUri)
      },
  { /* ignore */
    "Content-Base",
    HeaderInvalid,
    0,
    0
  },
  { /* ignore */
    "Content-Language",
    HeaderInvalid,
    0,
    0
  },
  { /* ignore */
    "Content-MD5",
    HeaderInvalid,
    0,
    0
  },
  {
    "Content-Range",
    HeaderContentDanger,
    0,
    0
  },
  {
    "Content-Type",
    HeaderInvalid,
    0,
    0
  },
  {
    "Content-Encoding",
    HeaderContentDanger,
    0,
    0
  },
  {0,HeaderInvalid,0,0}
};

const int numHeaderDescs=(sizeof(headerDescs)/sizeof(headerDescs[0]));

/* Read a header (just the Header: part, not the rest of the line) from
 *  the socket; returns the Header enum.
 */
Header getHeader(int sock)
{
  static const char** headers=0;
  int headerNum;

  if (!headers)
    {
      int i;

      headers=(const char**)malloc(sizeof(const char*)*numHeaderDescs);
      if (!headers)
        reportError("getHeader: malloc",8);

      for (i=0; i<numHeaderDescs; i++)
        headers[i]=headerDescs[i].name;
    }

  headerNum=expect(sock,headers,':',0);
  if (headerNum<0)
    return ((headerNum==-2) ? HeaderTerminator : HeaderInvalid);

  return headerDescs[headerNum].tag;
};

/* Read a header (the whole line) from the socket; returns -1 if we've
 *  reached the end of the headers, -3 if we've got a Content-* we
 *  can't handle, -2 if we can't read the header line, 0 otherwise.
 */
int readHeader(int sock)
{
  Header h=getHeader(sock);
  if (h==HeaderInvalid)
    return skipLine(sock);
  if (h==HeaderTerminator)
    return -1;
  if (h==HeaderContentDanger)
    return -3;

  if (readLine(sock,
               headerDescs[((int)h)-1].string,
               headerDescs[((int)h)-1].maxStringLen,
               1)
      <0
      )
    {
      (headerDescs[((int)h)-1].string)[0]=0;
      return -2;
    }
  else
    return 0;
}

/* Convenience function: write the given string to the socket */
int writeString(int sock,
                const char* str)
{
  return write(sock,str,strlen(str));
}

/* Convenience function: write the given integer to the socket */
int writeInt(int sock,
             int i)
{
  char buff[40];
  sprintf(buff,"%d",i);
  return writeString(sock,buff);
}

/* Look up the text message associated with the given HTTP status code.
 *  Doesn't know all the codes, just the ones we use.
 */
const char* getHttpStatusMsg(int status)
{
  static struct {
    int status;
    const char* msg;
  } statuses[]={
    {200,"OK"},
    {204,"No Content"},
    {304,"Not Modified"},
    {401,"Unauthorized"},
    {403,"Forbidden"},
    {404,"Not Found"},
    {500,"Internal Server Error"},
    {501,"Not Implemented"},
    {0,0}
  };

  int i;
  for (i=0; statuses[i].msg; i++)
    if (statuses[i].status==status)
      return statuses[i].msg;

  return "";
}

/* Send an HTTP status line (with the given status) and the header lines
 *  we'll always want to send.  Does not send the blank line terminating
 *  the headers.
 */
void sendHttpReplyHeaders(int sock,
                          int status)
{
  writeString(sock,"HTTP/1.0 ");
  {
    char statusStr[40];
    sprintf(statusStr,"%d",status);
    writeString(sock,statusStr);
  }

  writeString(sock," ");
  writeString(sock,getHttpStatusMsg(status));
  writeString(sock,"\n");
  writeString(sock,"Server: roamd/0.1\n");
  if (status!=404)
    writeString(sock,"Allow: GET, PUT, HEAD, DELETE, MOVE\n");
  writeString(sock,"Connection: close\n");
  {
    const char* dateStr=makeHttpTime(time(0));
    if (dateStr)
      {
        writeString(sock,"Date: ");
        writeString(sock,dateStr);
        writeString(sock,"\n");
      }
  }
}

/* Convenience wrapper around sendHttpReplyHeaders(): sends the
 *  given status and the terminating blank line.
 */
void sendHttpError(int sock,
                   int status)
{
  sendHttpReplyHeaders(sock,status);
  writeString(sock,"\n");
}

/* Convert an errno number into an HTTP status code */
int errnoToStatus(int err)
{
  switch (err)
    {
    case EISDIR:
    case EACCES:
    case EPERM:
    case EINVAL: /* Moving directory into a subdir of itself--can't happen */
    case EEXIST: /* Moving directory into a subdir of itself--can't happen */
    case EROFS:
      return 401;
      break;

    case ENOENT:
      return 404;
      break;

    case EXDEV: /* cross-file-system move--can't happen */
    case ENOTEMPTY:
    case EBUSY:
    case EMLINK: /* can't move because too many links on original */
    case ENOTDIR: /* Moving into a directory that isn't--can't happen */
    case EFAULT:
    case ENAMETOOLONG:
    case ENOMEM:
    case ELOOP:
    case ENOSPC:
    default:
      return 500;
      break;
    }
}

/* Report an HTTP error given an errno number */
void sendHttpErrorFromErrno(int sock,
                            int err)
{
  sendHttpError(sock,errnoToStatus(err));
}

/* Used by parseHttpTime(): given the given month name, return which
 *  month # it is.
 */
int monthNum(const char* month)
{
  const char* months[]={
    "Jan","Feb","Mar","Apr","May","Jun",
    "Jul","Aug","Sep","Oct","Nov","Dec"
  };

  int i;
  for (i=0; i<12; i++)
    {
      if (!strncasecmp(month,months[i],3))
        return i;
    }

  return 0;
}

/* Used by parseHttpTime(): convert the first n characters of str to an
 *  integer.  n cannot be greater than 7.
 *
 * Returns -1 on error.
 */
int natoi(const char* str,
          int n)
{
  char buff[8];

  if (n>(sizeof(buff)-1))
    return -1;

  if (!str)
    return -1;

  if (!(*str))
    return -1;

  if (n>strlen(str))
    return -1;

  {
    int i,j;
    for (i=j=0; i<n; i++)
      {
        if (!isspace(str[i]))
          buff[j++]=str[i];
      }
    buff[j]=0;
  }

  return atoi(buff);
}

/* Used by parseHttpTime(): parses out just the time part of the
 *  time/date spec.
 */
int fillTimePart(const char* time,
                 struct tm* t)
{
  int x;

  if (strlen(time)<8)
    return -1;

  if ((x=natoi(time,2))<0)
    return -1;
  t->tm_hour=x;

  if ((x=natoi(time+3,2))<0)
    return -1;
  t->tm_min=x;

  if ((x=natoi(time+6,2))<0)
    return -1;
  t->tm_sec=x;

  return 0;
}

/* Parse the given HTTP time string into a time_t */
time_t parseHttpTime(const char* timeStr)
{
  struct tm t;
  const char* space=strchr(timeStr,' ');

  if (!space)
    return 0;

  memset(&t,0,sizeof(t));

  {
    const char* date=space+1;
    if (!isdigit(*date))
      {
        /* asctime format */

        t.tm_mon=monthNum(date);
        space=strchr(date,' ');
        if (!space)
          return 0;

        {
          const char* mon=space+1;
          t.tm_mday=natoi(mon,3);
        
          if (t.tm_mday<0)
            return 0;

          space=strchr(mon,' ');
          if (!space)
            return 0;

          {
            const char* time=space+1;

            if (fillTimePart(time,&t)<0)
              return 0;

            space=strchr(time,' ');
            if (!space)
              return 0;

            {
              const char* year=space+1;
              t.tm_year=natoi(year,4);
              if (t.tm_year<0)
                return 0;
              t.tm_year-=1900;
            }
          }
        }
      }
    else /* not asctime */
      {
        t.tm_mday=natoi(date,2);

        if (t.tm_mday<0)
          return 0;

        if (date[2]=='-')
          {
            /* rfc850 format */

            if (strlen(date)<18)
              return 0;

            t.tm_mon=monthNum(date+3);
            if (t.tm_mon<0)
              return 0;

            t.tm_year=natoi(date+7,2);
            if (t.tm_year<0)
              return 0;

            if (fillTimePart(date+10,&t)<0)
              return 0;
          }
        else
          {
            /* rfc1123 format */
            if (strlen(date)<20)
              return 0;

            t.tm_mon=monthNum(date+3);
            if (t.tm_mon<0)
              return 0;

            t.tm_year=natoi(date+7,4);
            if (t.tm_year<0)
              return 0;
            t.tm_year-=1900;

            if (fillTimePart(date+12,&t)<0)
              return 0;
          }
      }
  }

  return mktime(&t);
}

/* Convert the given time_t into an HTTP time string */
const char* makeHttpTime(time_t t)
{
  struct tm* tm=gmtime(&t);
  static char buff[80];

  if (!tm)
    return 0;

  strftime(buff,80,"%a, %d %b %Y %H:%M:%S GMT",tm);
  return buff;
}

/* Perform a GET or HEAD (HEAD if headOnly is true) */
void doGet(int sock,
           const char* path, /* the pathname of the file to send */
           int headOnly)
{
  int fd=open(path,O_RDONLY);
  int length;
  const char* modTimeStr;

  if (fd<0)
    {
      sendHttpErrorFromErrno(sock,errno);
      return;
    }

  {
    struct stat buf;
    if (fstat(fd,&buf)<0)
      {
        sendHttpErrorFromErrno(sock,errno);
        close(fd);
        return;
      }
    length=buf.st_size;
    if (headerData.ifModifiedSince[0])
      {
        time_t ifModTime=parseHttpTime(headerData.ifModifiedSince);
        if (buf.st_mtime<=ifModTime)
          {
            sendHttpReplyHeaders(sock,304);
            writeString(sock,"\n");
            return;
          }
      }
    
    modTimeStr=makeHttpTime(buf.st_mtime);
    if (modTimeStr)
      modTimeStr=strdup(modTimeStr);
  }
  sendHttpReplyHeaders(sock,200);
  writeString(sock,"Content-Type: text/");
  writeString(sock,
              (strstr(path,"/bookmarks") ? "html" : "plain"));
  writeString(sock,"\nContent-Length: ");
  writeInt(sock,length);

  if (modTimeStr)
    {
      writeString(sock,"\nLast-Modified: ");
      writeString(sock,modTimeStr);
      free((void*)modTimeStr);
    }

  writeString(sock,"\n\n");

  if (!headOnly)
    {
      char buff[1024];
      int buffsize=sizeof(buff);
      int nbytes;
      while ((nbytes=read(fd,buff,buffsize))>0)
        {
          if (write(sock,buff,nbytes)<nbytes)
            {
#ifdef DEBUG
              fprintf(stderr,"Connection lost while sending file\n");
#endif
              break;
            }
        }
    }

  close(fd);
}

/* Build a pathname for a tmpfile that will receive an incoming PUT */
const char* getTmpPath(const char* path)
{
  char* tmpPath=malloc(strlen(path)+5);
  strcpy(tmpPath,path);
  strcat(tmpPath,".tmp");
  return tmpPath;
}

/* Perform a PUT method.  Writes the incoming data into a tmpfile;
 *  if everything succeeds, then it moves the tmpfile into the real target
 *  file.  This is to get the safety that Communicator would otherwise
 *  do by MOVE and PUT.
 */
void doPut(int sock,
           const char* path)
{
  int length=atoi(headerData.contentLength);
  int remaining=length;
  int alreadyExisted;
  const char* tmpPath=getTmpPath(path);
  int fd=open(tmpPath,O_WRONLY|O_CREAT|O_TRUNC,0600);
  if (fd<0)
    {
      int tmp=errno;
      sendHttpErrorFromErrno(sock,tmp);
      fprintf(stderr,"%s: can't open tmpfile %s: %s\n",
              progname,tmpPath,strerror(tmp));
      free((char*)tmpPath);
      return;
    }

  {
    char buff[1024];
    int bufsize=sizeof(buff);

    while (remaining>0)
      {
        int nbytes=(  (bufsize>remaining)
                    ? remaining
                    : bufsize
                    );
        int actual;
        int tmp;

#ifdef DEBUG
        printf("doPut(): about to read up to %d bytes (%d left)\n",
               nbytes,remaining);
#endif
        actual=read(sock,buff,nbytes);
        tmp=errno;
        if (actual<0)
          {
            sendHttpError(sock,500);
            fprintf(stderr,"%s: read failed (expected %d bytes): %s\n",
                    progname,
                    actual,nbytes,
                    strerror(tmp));
            free((char*)tmpPath);
            return;
          }

#ifdef DEBUG
        printf("doPut(): read %d bytes\n",actual);
#endif

        if (write(fd,buff,actual)<actual)
          {
            fprintf(stderr,"%s: can't write enough: %s\n",strerror(errno));
            sendHttpError(sock,500);
            free((char*)tmpPath);
            return;
          }

        remaining-=actual;

#ifdef DEBUG
        printf("doPut(): wrote %d bytes (%d left)\n",actual,remaining);
#endif

      }
  }

  close(fd);

  {
    struct stat buf;
    if (   (stat(path,&buf)<0)
        && (errno=ENOENT)
        )
      alreadyExisted=0;
    else
      alreadyExisted=1;
  }


  if (rename(tmpPath,path)<0)
    {
      int tmp=errno;
      sendHttpErrorFromErrno(sock,tmp);
      fprintf(stderr,"%s: can't rename file: %s\n",strerror(tmp));
      unlink(tmpPath);
      free((char*)tmpPath);
      return;
    }

  free((char*)tmpPath);

  sendHttpReplyHeaders(sock,(  alreadyExisted
                             ? (  (length>0)
                                ? 200
                                : 204
                                )
                             : 201
                             )
                       );
  writeString(sock,"\n");
}

/* Perform a DELETE method */
void doDelete(int sock,
              const char* path)
{
  if (unlink(path)<0)
    sendHttpErrorFromErrno(sock,errno);
  else
    sendHttpError(sock,204);
}

/* Perform a MOVE method */
void doMove(int sock,
            const char* path,
            const char* user)
{
  printf("MOVE %s to URI %s\n",path,headerData.newUri);
  if (strstr(headerData.newUri,"/li/")!=headerData.newUri)
    {
      fprintf(stderr,"%s: MOVE target URI (\"%s\") outside of /li/\n",
              progname,headerData.newUri);
      sendHttpError(sock,401);
      return;
    }

  {
    char userDirName[NAME_MAX+2];
    if (strlen(user)>NAME_MAX)
      {
        fprintf(stderr,"%s: MOVE username (\"%s\") too long\n",
                progname,user);
        sendHttpError(sock,404);
      }

    sprintf(userDirName,"%s/",user);
        
    if (strstr(headerData.newUri+4,userDirName)!=headerData.newUri+4)
      {
        fprintf(stderr,"%s: MOVE target URI (\"%s\") outside of /li/%s\n",
                progname,headerData.newUri,userDirName);
        sendHttpError(sock,401);
        return;
      }

    {
      const char* targetFile=headerData.newUri+4+strlen(userDirName);
      char targetPath[MAXPATHLEN];
      sprintf(targetPath,"%s/%s/%s",userDir,user,targetFile);
      if (rename(path,targetPath)<0)
        sendHttpErrorFromErrno(sock,errno);
      else
        sendHttpError(sock,200);
    }
  }
}

/* Process an incoming connection */
void processRequest(int sock)
{
  Method method=getMethod(sock);
  const char* user=0;
  const char* file=0;
  const char* path;

#ifdef DEBUG
  printf("Got method: %s\n",methodName(method));
#endif

  if (method==MethodInvalid)
    {
      sendHttpError(sock,501);
      return;
    }

  if (expect1(sock,"/li/")<0)
    {
#ifdef DEBUG
      fprintf(stderr,"Request-URI not rooted at /li/\n");
#endif
      sendHttpError(sock,404);
      return;
    }

#ifdef DEBUG
  printf("Got correct URI prefix\n");
#endif

  user=getUser(sock);
  if (!user)
    {
#ifdef DEBUG
      fprintf(stderr,"Request-URI contained illegal user\n");
#endif
      sendHttpError(sock,404);
      return;
    }
#ifdef DEBUG
  printf("Got user: %s\n",user);
#endif

  file=getFile(sock,tmpFilesLegal(method));
  if (!file)
    {
#ifdef DEBUG
      fprintf(stderr,"Request-URI contained illegal file\n");
#endif
      sendHttpError(sock,404);
      return;
    }
#ifdef DEBUG
  printf("Got file: %s\n",file);
#endif

  if (skipLine(sock)<0)
    {
#ifdef DEBUG
      fprintf(stderr,"Connection lost\n");
#endif
      return;
    }

  {
    int i;
    for (i=0; i<(numHeaderDescs-1); i++)
      {
        if (headerDescs[i].string)
          headerDescs[i].string[0]=0;
      }
  }

#ifdef DEBUG
  if (method==MethodMove)
    suppressExpectIllegals=0;
#endif

  {
    int headerResult;
    while ((headerResult=readHeader(sock))==0)
      ;

    if (headerResult==-2)
      {
#ifdef DEBUG
        fprintf(stderr,"Badly formatted header\n");
#endif
        return;
      }

    if (headerResult==-3)
      {
        sendHttpError(sock,501);
        return;
      }
  }

  
#ifdef DEBUG
  suppressExpectIllegals=1;
#endif


  {
    int i;
    for (i=0; i<(numHeaderDescs-1); i++)
      {
        if (headerDescs[i].string)
          {
            {
              char* cr=strchr(headerDescs[i].string,'\r');
              if (cr)
                (*cr)=0;
            }
#ifdef DEBUG
            printf("%s: %s\n",
                   headerDescs[i].name,
                   headerDescs[i].string);
#endif
          }
      }
  }

  if (authorize(user,headerData.authorization)!=1)
    {
      sendHttpReplyHeaders(sock,401);
      writeString(sock,"WWW-Authenticate: Basic realm=\"");
      writeString(sock,getRealm(user));
      writeString(sock,"\"\n\n");
      return;
    }

  path=getFilePath(user,file);

  switch (method)
    {
    case MethodGet:
    case MethodHead:
      doGet(sock,path,(method==MethodHead));
      break;

    case MethodPut:
      doPut(sock,path);
      break;

    case MethodMove:
      doMove(sock,path,user);
      break;

    default:
      sendHttpError(sock,501);
      break;
    }
}
